#include <iostream>
#include <vector>
#include <stack>
#include <iomanip>
#include <cstdlib>
#include <cmath>
#include "trng.h"
#include "fibonacci.h"

using namespace std;
using namespace TRNG;


typedef struct {
  int x;
  int y;
} koord;

class lattice {
private:
  vector< vector<int> > s;
  int L;
  int pos(int);
public:
  int size(void);
  void resize(int, int);
  void fill(int);
  void flipp(koord);
  int get(koord);
  void set(koord, int);
  double energy(void);
  double magnet(void);
  void print(void);
  lattice(int);
  lattice();
};

inline int lattice::pos(int x) {
  int t;
  return((t=x%L)<0 ? t+L : t);
}

int lattice::size(void) {
  return(L);
}

void lattice::resize(int newL, int w=-1) {
  int i;
  if (newL>0) {
    L=newL;
    s.resize(L);
    for (i=0; i<L; ++i)
      s[i].resize(L, w);
  } else {
    cerr << "negative size\n";
    exit(EXIT_FAILURE);
  }
}

void lattice::flipp(koord r) {
  s[pos(r.x)][pos(r.y)]*=-1;
}

void lattice::fill(int w) {
  int i, j;
  for (i=0; i<L; ++i)
    for(j=0; j<L; ++j)
      s[i][j]=w;
}

int lattice::get(koord r) {
  return(s[pos(r.x)][pos(r.y)]);
}

void lattice::set(koord r, int w) {
  s[pos(r.x)][pos(r.y)]=w;
}

double lattice::energy(void) {
  int i, j; 
  double e;
  e=0.0;
  for (i=0; i<L; ++i)
    for(j=0; j<L; ++j)
      e-=s[i][j]*(s[pos(i+1)][j]+s[i][pos(j+1)]);
  return(e/(s.size()*s.size()));
}

double lattice::magnet(void) {
  int i, j; 
  double m;
  m=0.0;
  for (i=0; i<L; ++i)
    for(j=0; j<L; ++j)
      if (s[i][j]<0)
	--m;
      else
	++m;
  return(fabs(m)/(s.size()*s.size()));
}

void lattice::print(void) {
  int i, j;
  for (i=0; i<L; ++i) {
    for(j=0; j<L; ++j)
      if (s[i][j]<0)
	cout << '.';
      else
	cout << '#';
    cout << endl;
  }
  cout << endl;
}

lattice::lattice(int newL) {
  lattice::resize(newL);
}

lattice::lattice() {
  L=0;
}

void wolffstep(RNG &R, lattice &s, double T) {
  stack<koord> buffer;
  double padd;
  int oldspin;
  koord r;
  
  padd=1.0-exp(-2.0/T);
  r.x=static_cast<int>(s.size()*R.uniform());
  r.y=static_cast<int>(s.size()*R.uniform());
  oldspin=s.get(r);
  s.flipp(r);
  buffer.push(r);
  while(!buffer.empty()) {
    r=buffer.top();
    buffer.pop();
    --r.x;
    if (s.get(r)==oldspin)
      if (R.boolean(padd)) {
	buffer.push(r);
	s.flipp(r);
      }
    ++r.x;
    ++r.x;
    if (s.get(r)==oldspin)
      if (R.boolean(padd)) {
	buffer.push(r);
	s.flipp(r);
      }
    --r.x;
    --r.y;
    if (s.get(r)==oldspin)
      if (R.boolean(padd)) {
	buffer.push(r);
	s.flipp(r);
      }
    ++r.y;
    ++r.y;
    if (s.get(r)==oldspin)
      if (R.boolean(padd)) {
	buffer.push(r);
	s.flipp(r);
      }
    --r.y;
  }
}

void output(vector<double> &Ea, vector<double> &ca, int simulations, 
	    double E_exact, double c_exact) {
  int j;
  Ea[simulations]=0.0;
  ca[simulations]=0.0;
  Ea[simulations+1]=0.0;
  ca[simulations+1]=0.0;
  for (j=0; j<simulations; ++j) {
    Ea[simulations]+=Ea[j]/simulations;
    ca[simulations]+=ca[j]/simulations;
  }
  for (j=0; j<simulations; ++j) {
    Ea[simulations+1]+=(Ea[j]-Ea[simulations])*(Ea[j]-Ea[simulations]);
    ca[simulations+1]+=(ca[j]-ca[simulations])*(ca[j]-ca[simulations]);
  }
  Ea[simulations+1]/=simulations-1.0;
  Ea[simulations+1]=sqrt(Ea[simulations+1]);
  ca[simulations+1]/=simulations-1.0;
  ca[simulations+1]=sqrt(ca[simulations+1]);
  cout << endl << "\t E\t\t c" << endl;
  for (j=0; j<simulations; ++j)
    cout << '\t' << setprecision(8) << Ea[j] << '\t' 
	 << setprecision(8) << ca[j] << endl;
  cout << "\t--------------\t--------------" << endl;
  cout << "mean\t" << setprecision(8) << Ea[simulations] << '\t' 
       << setprecision(8) << ca[simulations] << endl;
  cout << "Del\t" 
       << setprecision(8) << Ea[simulations]-E_exact << '\t' 
       << setprecision(8) << ca[simulations]-c_exact << endl;
  cout << "sig\t" << setprecision(8) << Ea[simulations+1] << '\t' 
       << setprecision(8) << ca[simulations+1] << endl;
  cout << "Del/sig\t" 
       << setprecision(8) 
       << fabs(Ea[simulations]-E_exact)/Ea[simulations+1] << '\t' 
       << setprecision(8) 
       << fabs(ca[simulations]-c_exact)/ca[simulations+1] << endl;
}


void wolff_main(RNG &R) {
  const int L=16;
  const double E_exact=-1.453064854; // 16x16
  const double c_exact=1.498704952;  // 16x16
  //  const double E_exact=-1.465960817; // 12x12
  //  const double c_exact= 1.352950680; // 12x12
  //  const double E_exact=-1.491589107;  // 8x8
  //  const double c_exact= 1.145559238;  // 8x8
  const int runs=1000000;
  const int simulations=10;
  int i, j;
  double T, E, E2, q, c;
  vector<double> Ea(simulations+2);
  vector<double> ca(simulations+2);

  cout.setf(ios::fixed);
  cout.setf(ios::showpoint);
  T=2.0/log(1+sqrt(2.0));
  lattice s(L);
  s.fill(-1);
  cout << R.name() << endl;
  cout << "T = " << T << endl;
  for (j=0; j<simulations; ++j) { 
    E=0.0;
    E2=0.0;
    for (i=0; i<runs; ++i) {
      wolffstep(R, s, T);
      q=s.energy();
      E+=q;
      E2+=q*q;
    }
    E/=runs;
    E2/=runs;
    c=(L*L)/(T*T)*(E2-E*E);
    Ea[j]=E;
    ca[j]=c;
  }
  output(Ea, ca, simulations, E_exact, c_exact);
  cout << endl 
       << "--------------------------------------" << endl
       << endl;
}

int main (void) {
  Fibonacci r1(103, 250);
  wolff_main(r1);
  return(EXIT_SUCCESS);
}





